""" Exception related objects and helpers."""

class Error(Exception):
    """ Base Error for all gsh exceptions."""


class ConfigError(Error):
    """ Error parsing configuration."""

class LoaderError(Error):
    """ Error loading hosts from loader plugin."""

class EarlyExit(Error):
    """ Used to bail out of hooks during the pre_job stage."""



import gevent
from gevent.queue import Queue, Empty
from gevent_subprocess import Popen, PIPE

from gsh.plugin import BaseExecutor, BaseInnerExecutor

class SshExecutor(BaseExecutor):
    class Executor(BaseInnerExecutor):
        def __init__(self, *args, **kwargs):
            self.names = {}
            self._output_queue = Queue()
            super(SshExecutor.Executor, self).__init__(*args, **kwargs)

        @staticmethod
        def _stream_fd(fd, queue):
            for line in iter(fd.readline, b""):
                queue.put_nowait((fd, line))

        def _consume(self, queue):
            while True:
                try:
                    output = queue.get()
                except Empty:
                    continue

                # None is explicitly sent to shutdown the consumer
                if output is None:
                    return

                fd, line = output
                self.update(self.hostname, self.names[fd], line)

        def run(self):
            _proc = Popen(
                ["ssh", "-o", "PasswordAuthentication=no", self.hostname] + self.command,
                stdout=PIPE, stderr=PIPE
            )

            self.names = {
                _proc.stdout: "stdout",
                _proc.stderr: "stderr",
            }

            out_worker = gevent.spawn(self._stream_fd, _proc.stdout, self._output_queue)
            err_worker = gevent.spawn(self._stream_fd, _proc.stderr, self._output_queue)
            waiter = gevent.spawn(_proc.wait)
            consumer = gevent.spawn(self._consume, self._output_queue)

            gevent.joinall([out_worker, err_worker, waiter], timeout=self.timeout)

            # If we've made it here and the process hasn't completed we've timed out.
            if _proc.poll() is None:
                self._output_queue.put_nowait(
                    (_proc.stderr, "GSH: command timed out after %s second(s).\n" % self.timeout))
                _proc.kill()

            rc = _proc.wait()

            self._output_queue.put_nowait(None)
            consumer.join()

            return rc




import gevent
from gevent.pool import Pool
import time

from .plugin import get_executors
from .exceptions import EarlyExit


class RemotePopen(object):

    QUEUED = 0
    RUNNING = 1
    FAILED = 2
    SUCCESS = 3

    def __init__(self, hostname, command, timeout=None, hooks=None, executor=None):
        self.hostname = hostname
        self.command = command
        self.timeout = timeout

        if executor is None:
            executor = get_executors().SshExecutor()

        self.executor = executor.Executor(
            executor, self.hostname, self.command, self.timeout,
            self._run_update_host_hooks
        )

        # Treat 0 second timeouts as no timeout.
        if not timeout:
            self.timeout = None

        if hooks is None:
            hooks = []
        self.hooks = hooks

        self.status = RemotePopen.QUEUED
        self.rc = None


        self._pre_host_hooks = None
        self._post_host_hooks = None

    def _run_pre_host_hooks(self):
        for hook in self.hooks:
            hook.pre_host(self.hostname, time.time())

    def _run_post_host_hooks(self):
        for hook in self.hooks:
            hook.post_host(self.hostname, self.rc, time.time())

    def _run_update_host_hooks(self, hostname, stream, line):
        for hook in self.hooks:
            hook.update_host(hostname, stream, line)

    def run(self):
        self._pre_host_hooks = gevent.spawn(self._run_pre_host_hooks)
        self._pre_host_hooks.join()

        self.status = RemotePopen.RUNNING

        self.rc = self.executor.run()

        if not self.rc:
            self.status = RemotePopen.SUCCESS
        else:
            self.status = RemotePopen.FAILED

        self._post_host_hooks = gevent.spawn(self._run_post_host_hooks)
        self._post_host_hooks.join()


class Gsh(object):
    def __init__(self, hosts, command, fork_limit=1, timeout=None, hooks=None, executor=None):
        self.hosts = set(hosts)
        self.command = command
        self.fork_limit = self._build_fork_limit(fork_limit, len(self.hosts))
        self.timeout = timeout

        self.executor = executor
        if executor is None:
            self.executor = get_executors().SshExecutor()

        # Treat 0 second timeouts as no timeout.
        if not timeout:
            self.timeout = None

        if hooks is None:
            hooks = []
        self.hooks = hooks

        self._pool = Pool(max(self.fork_limit, 1))
        self._greenlets = []
        self._remotes = []

        self._pre_job_hooks = None
        self._post_job_hooks = None

        self._continue = True

    @staticmethod
    def _build_fork_limit(fork_limit, num_hosts):
        if isinstance(fork_limit, int) or fork_limit.isdigit():
            return int(fork_limit)
        if fork_limit.endswith("%"):
            return int(float(num_hosts) * (float(fork_limit[:-1]) / 100.0))
        # If we can't parse your forklimit go serial for safety.
        return 1

    def run_async(self):

        # Don't start executing until the pre_job hooks have completed.
        self._pre_job_hooks = gevent.spawn(self._run_pre_job_hooks)
        self._pre_job_hooks.join()

        if self._continue:
            for host in self.hosts:
                remote_command = RemotePopen(
                    host, self.command, hooks=self.hooks,
                    timeout=self.timeout, executor=self.executor)
                self._remotes.append(remote_command)
                self._greenlets.append(self._pool.apply_async(remote_command.run))

        self._post_job_hooks = gevent.spawn(self._run_post_job_hooks)

    def _run_pre_job_hooks(self):
        for hook in self.hooks:
            try:
                hook.pre_job(self.command, self.hosts, time.time())
            except EarlyExit:
                self._continue = False

    def _run_post_job_hooks(self):
        # Wait for all greenlets to finish before running these hooks.
        gevent.joinall(self._greenlets)
        for hook in self.hooks:
            hook.post_job(time.time())

    def wait(self, timeout=None):
        rc = 0
        gevent.joinall(self._greenlets + [self._post_job_hooks], timeout=timeout, raise_error=True)
        for remote in self._remotes:
            if remote.rc:
                return remote.rc
        return rc




import sys
import math
from gsh.plugin import BaseExecutionHook


class PrinterHook(BaseExecutionHook):
    def __init__(self, *args, **kwargs):
        self.prepend_host = kwargs.pop("prepend_host", True)
        self.show_percent = kwargs.pop("show_percent", False)
        self.add_newline = kwargs.pop("add_newline", True)

        self.hosts_total = 0
        self.hosts_finished = 0
        self.hosts_percent = 0
        self.longest_len = 0

        super(PrinterHook, self).__init__(*args, **kwargs)

    def pre_job(self, command, hosts, timestamp):
        self.hosts_total = len(hosts)
        if hosts:
            # Add one to account for colons
            self.longest_len = len(max(hosts, key=len)) + 1

    def update_host(self, hostname, stream, line):
        if stream not in ("stdout", "stderr"):
            return
        writer = getattr(sys, stream)
        line = line.decode('utf-8')

        try:
            self._write(writer, hostname, line)
        except IOError as err:
            if writer is not sys.stderr:
                output = "Failed to write to stream %s: (%s, %s)\n" % (stream, err.errno, err.strerror)
                self._write(sys.stderr, hostname, output)

    def post_host(self, hostname, return_code, timestamp):
        self.hosts_finished += 1
        self.hosts_percent = int(math.ceil(float(self.hosts_finished) * 100 / self.hosts_total))

    def _write(self, stream, hostname, line):
        if self.prepend_host:
            line = "%%-%ss %%s" % self.longest_len % (hostname + ":", line)

        if self.show_percent:
            line = "(%3s%%) %s" % (self.hosts_percent, line)

        if self.add_newline and line and line[-1] != "\n":
            line += "\n"

        line = line.encode('utf-8')
        stream.write(line)
        stream.flush()



import getpass
import gevent
import threading

from gsh.plugin import BaseExecutor, BaseInnerExecutor


class _PexpectThread(threading.Thread):
    def __init__(self, executor):
        self.executor = executor
        super(_PexpectThread, self).__init__()

    def run(self):
        try:
            # Defer import since most people won't want to use this executor
            # and I don't want it to be required to use gsh.
            import pxssh

            ssh = pxssh.pxssh()
            ssh.force_password = True
            ssh.login(self.executor.hostname, self.executor.parent.username, self.executor.parent.password)


            command = " ".join(self.executor.command)
            num_commands = command.count("\n") + 1
            num_prompts = 0
            ssh.sendline(command)

            # We have to block for some amount of time to wait for the data to be transfered.
            while True:
                if num_prompts >= num_commands:
                    break

                num_prompts += 1
                ssh.prompt()

                for line in ssh.before.splitlines()[1:]:
                    if not line:
                        continue
                    self.executor.update(self.executor.hostname, "stdout", line)

            ssh.logout()
        except pxssh.ExceptionPxssh, err:
            self.executor.update(self.executor.hostname, "stderr", "GSH: Failed to login: %s" % err)
            self.executor.rv = 1
            return


class PexpectExecutor(BaseExecutor):
    def __init__(self, args, kwargs):
        # This import is unused but I want it to blow up
        # at the job level if it's not found
        import pxssh

        super(PexpectExecutor, self).__init__(args, kwargs)
        self.password = getpass.getpass("Password: ")
        self.username = getpass.getuser()

    class Executor(BaseInnerExecutor):
        def __init__(self, *args, **kwargs):
            super(PexpectExecutor.Executor, self).__init__(*args, **kwargs)
            self.rv = 0

        def run(self):

            thread = _PexpectThread(self)
            thread.daemon = True
            thread.start()

            sleep_duration = 0.01
            while True:
                alive = thread.is_alive()
                if not alive:
                    return self.rv
                gevent.sleep(sleep_duration)
                if sleep_duration < 0.5:
                    sleep_duration *= 2

            return self.rv





import sys
from gsh.plugin import BaseExecutionHook

class DebuggerHook(BaseExecutionHook):
    def update_host(self, hostname, stream, line):
        print "Running update_host hook: ", (hostname, stream, line)

    def pre_host(self, *args):
        print "Running pre_host hook:", args

    def post_host(self, *args):
        print "Running post_host hook:", args

    def pre_job(self, *args):
        print "Running pre_job hook:", args

    def post_job(self, *args):
        print "Running post_job hook:", args




import logging

from gsh.plugin import BaseHostLoader

class FileLoader(BaseHostLoader):
    opt_short = "-f"
    opt_long = "--file"
    opt_metavar = "FILE"
    opt_help = "Get a list of machines from the specified file."

    def __call__(self, *args):
        hosts = []
        for host_file in args:
            hosts.extend([line.strip() for line in open(host_file)])
        return hosts

